@@ -14,6 +14,7 @@ NULL
 #' ```
 #' p <- process$new(command = NULL, args,
 #'                  stdin = NULL, stdout = NULL, stderr = NULL,
+#'                  pty = FALSE, pty_options = list(),
 #'                  connections = list(), poll_connection = NULL,
 #'                  env = NULL, cleanup = TRUE, cleanup_tree = FALSE,
 #'                  wd = NULL, echo_cmd = FALSE, supervise = FALSE,
@@ -24,8 +25,8 @@ NULL
 #' p$is_alive()
 #' p$signal(signal)
 #' p$interrupt()
-#' p$kill(grace = 0.1)
-#' p$kill_tree(grace = 0.1)
+#' p$kill(grace = 0.1, close_connections = TRUE)
+#' p$kill_tree(grace = 0.1, close_connections = TRUE)
 #' p$wait(timeout = -1)
 #' p$get_pid()
 #' p$get_exit_status()
@@ -85,14 +86,32 @@ NULL
 #'     used as is, without a shell. They don't need to be escaped.
 #' * `stdin`: What to do with the standard input. Possible values:
 #'     `NULL`: set to the _null device_, i.e. no standard input is
-#'     provided; a string, supply the specified string as standard input;
+#'     provided; a file name, use this file as standard input;
 #'     `"|"`: create a (writeable) connection for stdin.
 #' * `stdout`: What to do with the standard output. Possible values:
 #'     `NULL`: discard it; a string, redirect it to this file;
 #'     `"|"`: create a connection for it.
 #' * `stderr`: What to do with the standard error. Possible values:
 #'     `NULL`: discard it; a string, redirect it to this file;
-#'     `"|"`: create a connection for it.
+#'     `"|"`: create a connection for it; `"2>&1"`: redirect it to the
+#'     same connection (i.e. pipe or file) as `stdout`. `"2>&1"` is a
+#'     way to keep standard output and error correctly interleaved.
+#' * `pty`: Whether to create a pseudo terminal (pty) for the background
+#'     process. This is currently only supported on Unix systems, but
+#'     not supported on Solaris.
+#'     If it is `TRUE`, then the `stdin`, `stdout` and `stderr` arguments
+#'     must be `NULL`. If a pseudo terminal is created, then processx
+#'     will create pipes for standard input and standard output. There is
+#'     no separate pipe for standard error, because there is no way to
+#'     distinguish between stdout and stderr on a pty. Note that the
+#'     standard output connection of the pty is _blocking_, so we always
+#'     poll the standard output connection before reading from it using
+#'     the `$read_output()` method. Also, because `$read_output_lines()`
+#'     could still block if no complete line is available, this function
+#'     always fails if the process has a pty. Use `$read_output()` to
+#'     read from ptys.
+#' * `pty_options`: Unix pseudo terminal options, a named list. see
+#'     [default_pty_options()] for details and defaults.
 #' * `connections`: A list of connections to pass to the child process.
 #'     This is an experimental feature currently.
 #' * `poll_connection`: Whether to create an extra connection to the process
@@ -128,6 +147,9 @@ NULL
 #' * `signal`: An integer scalar, the id of the signal to send to
 #'     the process. See [tools::pskill()] for the list of signals.
 #' * `grace`: Currently not used.
+#' * `close_connections`: Whether to close standard input, standard output,
+#'    standard error connections and the poll connection, after killing
+#'    the process.
 #' * `timeout`: Timeout in milliseconds, for the wait or the I/O
 #'     polling.
 #' * `n`: Number of characters or lines to read.
@@ -356,21 +378,45 @@ NULL
 #' can poll the output of several processes, and returns as soon as any
 #' of them has generated output (or exited).
 #'
+#' @section Cleaning up background processes:
+#' processx kills processes that are not referenced any more (if `cleanup`
+#' is set to `TRUE`), or the whole subprocess tree (if `cleanup_tree` is
+#' also set to `TRUE`).
+#'
+#' The cleanup happens when the references of the processes object are
+#' garbage collected. To clean up earlier, you can call the `kill()` or
+#' `kill_tree()` method of the process(es), from an `on.exit()` expression,
+#' or an error handler:
+#' ```r
+#' process_manager <- function() {
+#'   on.exit({
+#'     try(p1$kill(), silent = TRUE)
+#'     try(p2$kill(), silent = TRUE)
+#'   }, add = TRUE)
+#'   p1 <- process$new("sleep", "3")
+#'   p2 <- process$new("sleep", "10")
+#'   p1$wait()
+#'   p2$wait()
+#' }
+#' process_manager()
+#' ```
+#'
+#' If you interrupt `process_manager()` or an error happens then both `p1`
+#' and `p2` are cleaned up immediately. Their connections will also be
+#' closed. The same happens at a regular exit.
+#'
 #' @name process
-#' @examples
-#' # CRAN does not like long-running examples
-#' \dontrun{
+#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
 #' p <- process$new("sleep", "2")
 #' p$is_alive()
 #' p
 #' p$kill()
 #' p$is_alive()
 #'
-#' p <- process$new("sleep", "2")
+#' p <- process$new("sleep", "1")
 #' p$is_alive()
-#' Sys.sleep(3)
+#' Sys.sleep(2)
 #' p$is_alive()
-#' }
 #'
 NULL
 
@@ -385,16 +431,16 @@ process <- R6::R6Class(
   public = list(
 
     initialize = function(command = NULL, args = character(),
-      stdin = NULL, stdout = NULL, stderr = NULL, connections = list(),
-      poll_connection = NULL, env = NULL, cleanup = TRUE,
-      cleanup_tree = FALSE, wd = NULL, echo_cmd = FALSE, supervise = FALSE,
-      windows_verbatim_args = FALSE, windows_hide_window = FALSE,
-      encoding = "",  post_process = NULL)
+      stdin = NULL, stdout = NULL, stderr = NULL, pty = FALSE,
+      pty_options = list(), connections = list(), poll_connection = NULL,
+      env = NULL, cleanup = TRUE, cleanup_tree = FALSE, wd = NULL,
+      echo_cmd = FALSE, supervise = FALSE, windows_verbatim_args = FALSE,
+      windows_hide_window = FALSE, encoding = "",  post_process = NULL)
 
       process_initialize(self, private, command, args, stdin,
-                         stdout, stderr, connections, poll_connection,
-                         env, cleanup, cleanup_tree, wd, echo_cmd,
-                         supervise, windows_verbatim_args,
+                         stdout, stderr, pty, pty_options, connections,
+                         poll_connection, env, cleanup, cleanup_tree, wd,
+                         echo_cmd, supervise, windows_verbatim_args,
                          windows_hide_window, encoding, post_process),
 
     finalize = function() {
@@ -402,11 +448,11 @@ process <- R6::R6Class(
           ps::ps_is_supported()) self$kill_tree()
     },
 
-    kill = function(grace = 0.1)
-      process_kill(self, private, grace),
+    kill = function(grace = 0.1, close_connections = TRUE)
+      process_kill(self, private, grace, close_connections),
 
-    kill_tree = function(grace = 0.1)
-      process_kill_tree(self, private, grace),
+    kill_tree = function(grace = 0.1, close_connections = TRUE)
+      process_kill_tree(self, private, grace, close_connections),
 
     signal = function(signal)
       process_signal(self, private, signal),
@@ -558,6 +604,8 @@ process <- R6::R6Class(
     stdin = NULL,         # stdin argument or stream
     stdout = NULL,        # stdout argument or stream
     stderr = NULL,        # stderr argument or stream
+    pty = NULL,           # whether we should create a PTY
+    pty_options = NULL,   # various PTY options
     pstdin = NULL,        # the original stdin argument
     pstdout = NULL,       # the original stdout argument
     pstderr = NULL,       # the original stderr argument
@@ -590,7 +638,9 @@ process <- R6::R6Class(
     tree_id = NULL,
 
     get_short_name = function()
-      process_get_short_name(self, private)
+      process_get_short_name(self, private),
+    close_connections = function()
+      process_close_connections(self, private)
   )
 )
 
@@ -599,23 +649,29 @@ process <- R6::R6Class(
 
 process_wait <- function(self, private, timeout) {
   "!DEBUG process_wait `private$get_short_name()`"
-  .Call(c_processx_wait, private$status, as.integer(timeout))
+  rethrow_call_with_cleanup(
+    c_processx_wait, private$status,
+    as.integer(timeout),
+    private$get_short_name()
+  )
   invisible(self)
 }
 
 process_is_alive <- function(self, private) {
   "!DEBUG process_is_alive `private$get_short_name()`"
-  .Call(c_processx_is_alive, private$status)
+  rethrow_call(c_processx_is_alive, private$status, private$get_short_name())
 }
 
 process_get_exit_status <- function(self, private) {
   "!DEBUG process_get_exit_status `private$get_short_name()`"
-  .Call(c_processx_get_exit_status, private$status)
+  rethrow_call(c_processx_get_exit_status, private$status,
+               private$get_short_name())
 }
 
 process_signal <- function(self, private, signal) {
   "!DEBUG process_signal `private$get_short_name()` `signal`"
-  .Call(c_processx_signal, private$status, as.integer(signal))
+  rethrow_call(c_processx_signal, private$status, as.integer(signal),
+               private$get_short_name())
 }
 
 process_interrupt <- function(self, private) {
@@ -625,27 +681,28 @@ process_interrupt <- function(self, private) {
     st <- run(get_tool("interrupt"), c(pid, "c"), error_on_status = FALSE)
     if (st$status == 0) TRUE else FALSE
   } else {
-    .Call(c_processx_interrupt, private$status)
+    rethrow_call(c_processx_interrupt, private$status,
+                 private$get_short_name())
   }
 }
 
-process_kill <- function(self, private, grace) {
+process_kill <- function(self, private, grace, close_connections) {
   "!DEBUG process_kill '`private$get_short_name()`', pid `self$get_pid()`"
-  ret <- .Call(c_processx_kill, private$status, as.numeric(grace))
-  if (!is.null(p <- private$poll_pipe)) .Call(c_processx_connection_close, p)
+  ret <- rethrow_call(c_processx_kill, private$status, as.numeric(grace),
+                      private$get_short_name())
+  if (close_connections) private$close_connections()
   ret
 }
 
-process_kill_tree <- function(self, private, grace) {
+process_kill_tree <- function(self, private, grace, close_connections) {
   "!DEBUG process_kill_tree '`private$get_short_name()`', pid `self$get_pid()`"
   if (!ps::ps_is_supported()) {
-    stop(structure(
-      list(message = "kill_tree is not supported on this platform"),
-      class = c("not_implemented", "error", "condition")))
+    throw(new_not_implemented_error(
+      "kill_tree is not supported on this platform"))
   }
 
   ret <- get("ps_kill_tree", asNamespace("ps"))(private$tree_id)
-  if (!is.null(p <- private$poll_pipe)) .Call(c_processx_connection_close, p)
+  if (close_connections) private$close_connections()
   ret
 }
 
@@ -654,7 +711,7 @@ process_get_start_time <- function(self, private) {
 }
 
 process_get_pid <- function(self, private) {
-  .Call(c_processx_get_pid, private$status)
+  rethrow_call(c_processx_get_pid, private$status)
 }
 
 process_is_supervised <- function(self, private) {
@@ -673,7 +730,7 @@ process_supervise <- function(self, private, status) {
 }
 
 process_get_result <- function(self, private) {
-  if (self$is_alive()) stop("Process is still alive")
+  if (self$is_alive()) throw(new_error("Process is still alive"))
   if (!private$post_process_done && is.function(private$post_process)) {
     private$post_process_result <- private$post_process()
     private$post_process_done <- TRUE
@@ -688,3 +745,25 @@ process_as_ps_handle <- function(self, private) {
 ps_method <- function(fun, self) {
   fun(ps::ps_handle(self$get_pid(), self$get_start_time()))
 }
+
+process_close_connections <- function(self, private) {
+  for (f in c("stdin_pipe", "stdout_pipe", "stderr_pipe", "poll_pipe")) {
+    if (!is.null(p <- private[[f]])) {
+      rethrow_call(c_processx_connection_close, p)
+    }
+  }
+}
+
+#' Default options for pseudo terminals (ptys)
+#'
+#' @return Named list of default values of pty options.
+#'
+#' Options and default values:
+#' * `echo` whether to keep the echo on the terminal. `FALSE` turns echo
+#'   off.
+#'
+#' @export
+
+default_pty_options <- function() {
+  list(echo = FALSE)
+}
