@@ -1,18 +1,26 @@
 
 #ifndef _WIN32
 
-#include "../processx.h"
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE 1
+#endif
 
+#include <stdlib.h>
+#include <fcntl.h>
 #include <stdio.h>
 
+#include "../processx.h"
+#include "../cleancall.h"
+
 /* Internals */
 
 static void processx__child_init(processx_handle_t *handle, int (*pipes)[2],
 				 int stdio_count, char *command, char **args,
 				 int error_fd, const char *std_in,
 				 const char *std_out,
-				 const char *std_err, char **env,
-				 processx_options_t *options,
+				 const char *std_err,
+                                 const char *pty_name, char **env,
+                                 processx_options_t *options,
 				 const char *tree_id);
 
 static SEXP processx__make_handle(SEXP private, int cleanup);
@@ -42,11 +50,16 @@ void processx__create_connections(processx_handle_t *handle, SEXP private,
 extern char **environ;
 #endif
 
+#include <termios.h>
+#include <sys/ioctl.h>
+
 extern processx__child_list_t child_list_head;
 extern processx__child_list_t *child_list;
 extern processx__child_list_t child_free_list_head;
 extern processx__child_list_t *child_free_list;
 
+extern int processx__notify_old_sigchld_handler;
+
 /* We are trying to make sure that the variables in the library are
    properly set to their initial values after a library (re)load.
    This function is called from `R_init_processx`. */
@@ -61,6 +74,50 @@ void R_init_processx_unix() {
   child_free_list_head.weak_status = R_NilValue;
   child_free_list_head.next = 0;
   child_free_list = &child_free_list_head;
+
+  if (getenv("PROCESSX_NOTIFY_OLD_SIGCHLD")) {
+    processx__notify_old_sigchld_handler = 1;
+  }
+}
+
+int processx__pty_master_open(char *slave_name, size_t sn_len) {
+  int master_fd, saved_errno;
+  char *p;
+
+  master_fd = posix_openpt(O_RDWR | O_NOCTTY);
+  if (master_fd == -1) return -1;
+
+  if (grantpt(master_fd) == -1) {
+    saved_errno = errno;
+    close(master_fd);
+    errno = saved_errno;
+    return -1;
+  }
+
+  if (unlockpt(master_fd) == -1) {
+    saved_errno = errno;
+    close(master_fd);
+    errno = saved_errno;
+    return -1;
+  }
+
+  p = ptsname(master_fd);
+  if (p == NULL) {
+    saved_errno = errno;
+    close(master_fd);
+    errno = saved_errno;
+    return -1;
+  }
+
+  if (strlen(p) < sn_len) {
+    strncpy(slave_name, p, sn_len);
+  } else {
+    close(master_fd);
+    errno = EOVERFLOW;
+    return -1;
+  }
+
+  return master_fd;
 }
 
 /* These run in the child process, so no coverage here. */
@@ -75,21 +132,83 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
 				 int stdio_count, char *command, char **args,
 				 int error_fd, const char *std_in,
 				 const char *std_out,
-				 const char *std_err, char **env,
+				 const char *std_err,
+                                 const char *pty_name, char **env,
 				 processx_options_t *options,
 				 const char *tree_id) {
 
   int close_fd, use_fd, fd, i;
   const char *out_files[3] = { std_in, std_out, std_err };
+  int min_fd = 0;
 
   setsid();
 
-  /* The dup2 calls make sure that stdin, stdout and stderr use file
-     descriptors 0, 1 and 2 respectively. */
+  /* Do we need a pty? */
+  if (pty_name) {
+    /* Do not mess with stdin/stdout/stderr, all handled by the pty */
+    min_fd = 3;
 
-  for (fd = 0; fd < stdio_count; fd++) {
+    int slave_fd = open(pty_name, O_RDWR);
+    if (slave_fd == -1) {
+      processx__write_int(error_fd, -errno);
+      raise(SIGKILL);
+    }
+
+#ifdef TIOCSCTTY
+    if (ioctl(slave_fd, TIOCSCTTY, 0) == -1) {
+      processx__write_int(error_fd, -errno);
+      raise(SIGKILL);
+    }
+#endif
+
+    struct termios tp;
+
+    if (tcgetattr(slave_fd, &tp) == -1) {
+      processx__write_int(error_fd, -errno);
+      raise(SIGKILL);
+    }
+
+    if (options->pty_echo) {
+      tp.c_lflag |= ECHO;
+    } else {
+      tp.c_lflag &= ~ECHO;
+    }
+
+    if (tcsetattr(slave_fd, TCSAFLUSH, &tp) == -1) {
+      processx__write_int(error_fd, -errno);
+      raise(SIGKILL);
+    }
+
+    /* TODO: set other terminal attributes and size */
+
+    /* Duplicate pty slave to be child's stdin, stdout, and stderr */
+    if (dup2(slave_fd, STDIN_FILENO) != STDIN_FILENO) {
+      processx__write_int(error_fd, -errno);
+      raise(SIGKILL);
+    }
+    if (dup2(slave_fd, STDOUT_FILENO) != STDOUT_FILENO) {
+      processx__write_int(error_fd, -errno);
+      raise(SIGKILL);
+    }
+    if (dup2(slave_fd, STDERR_FILENO) != STDERR_FILENO) {
+      processx__write_int(error_fd, -errno);
+      raise(SIGKILL);
+    }
+
+    if (slave_fd > STDERR_FILENO) close(slave_fd);
+  }
+
+  /* We want to prevent use_fd < fd, because we will dup2() use_fd into
+     fd later. If use_fd >= fd, then this is always possible,
+     without mixing up stdin, stdout and stderr. Without this, we could
+     have a case when we dup2() 2 into 1, and then 1 is lost. */
+
+  for (fd = min_fd; fd < stdio_count; fd++) {
     use_fd = pipes[fd][1];
+    /* If use_fd < 0 then there is no pipe for fd. */
     if (use_fd < 0 || use_fd >= fd) continue;
+    /* If use_fd < fd, then we create a brand new fd for it,
+       starting at stdio_count, which is bigger then fd, surely. */
     pipes[fd][1] = fcntl(use_fd, F_DUPFD, stdio_count);
     if (pipes[fd][1] == -1) {
       processx__write_int(error_fd, -errno);
@@ -97,21 +216,41 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
     }
   }
 
-  for (fd = 0; fd < stdio_count; fd++) {
+  /* This loop initializes the stdin, stdout, stderr fds of the child
+     process properly. */
+
+  for (fd = min_fd; fd < stdio_count; fd++) {
+    /* close_fd is an fd that must be closed. Initially this is the
+       parent's end of a pipe. (-1 if no pipe for this fd.) */
     close_fd = pipes[fd][0];
+    /* use_fd is the fd that the child must use for stdin/out/err. */
     use_fd = pipes[fd][1];
 
-    if (use_fd < 0) {
+    /* If no pipe, then we see if this is the 2>&1 case. */
+    if (fd == 2 && use_fd < 0 && out_files[fd] &&
+	! strcmp(out_files[fd], "2>&1")) {
+      use_fd = 1;
+
+    } else if (use_fd < 0) {
+      /* Otherwise we open a file. If the stdin/out/err is not
+	 requested, then we open a file to /dev/null */
+      /* For fd >= 3, the fd is just passed, and we just use it,
+	 no need to open any file */
       if (fd >= 3) continue;
+
       if (out_files[fd]) {
+	/* A file was requested, open it */
 	if (fd == 0) {
 	  use_fd = open(out_files[fd], O_RDONLY);
 	} else {
 	  use_fd = open(out_files[fd], O_CREAT | O_TRUNC| O_RDWR, 0644);
 	}
       } else {
+	/* NULL, so stdin/out/err is ignored, using /dev/null */
 	use_fd = open("/dev/null", fd == 0 ? O_RDONLY : O_RDWR);
       }
+      /* In the output file case, we might need to close use_fd, after
+	 we dup2()-d it into fd. */
       close_fd = use_fd;
 
       if (use_fd == -1) {
@@ -120,6 +259,10 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
       }
     }
 
+    /* We will use use_fd for fd. If they happen to be equal, make
+       sure that fd is _not_ closed on exec. Otherwise dup2() use_fd
+       into fd. dup2() clears the CLOEXEC flag, so no need for a fcntl
+       call in this case. */
     if (fd == use_fd) {
       processx__cloexec_fcntl(use_fd, 0);
     } else {
@@ -133,10 +276,12 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
 
     if (fd <= 2) processx__nonblock_fcntl(fd, 0);
 
+    /* If we have an extra fd, that we already dup2()-d into fd,
+       we can close it now. */
     if (close_fd >= stdio_count) close(close_fd);
   }
 
-  for (fd = 0; fd < stdio_count; fd++) {
+  for (fd = min_fd; fd < stdio_count; fd++) {
     use_fd = pipes[fd][1];
     if (use_fd >= stdio_count) close(use_fd);
   }
@@ -218,7 +363,7 @@ static SEXP processx__make_handle(SEXP private, int cleanup) {
   SEXP result;
 
   handle = (processx_handle_t*) malloc(sizeof(processx_handle_t));
-  if (!handle) { error("Out of memory"); }
+  if (!handle) { R_THROW_ERROR("Cannot make processx handle, out of memory"); }
   memset(handle, 0, sizeof(processx_handle_t));
   handle->waitpipe[0] = handle->waitpipe[1] = -1;
 
@@ -235,7 +380,7 @@ static void processx__handle_destroy(processx_handle_t *handle) {
   free(handle);
 }
 
-void processx__make_socketpair(int pipe[2]) {
+void processx__make_socketpair(int pipe[2], const char *exe) {
 #if defined(__linux__)
   static int no_cloexec;
   if (no_cloexec)  goto skip;
@@ -247,7 +392,7 @@ void processx__make_socketpair(int pipe[2]) {
    * Anything else is a genuine error.
    */
   if (errno != EINVAL) {
-    error("processx socketpair: %s", strerror(errno)); /* LCOV_EXCL_LINE */
+    R_THROW_SYSTEM_ERROR("processx socketpair");
   }
 
   no_cloexec = 1;
@@ -256,7 +401,12 @@ skip:
 #endif
 
   if (socketpair(AF_UNIX, SOCK_STREAM, 0, pipe)) {
-    error("processx socketpair: %s", strerror(errno));
+    if (exe) {
+      R_THROW_SYSTEM_ERROR("cannot make processx socketpair while "
+                           "running '%s'", exe);
+    } else {
+      R_THROW_SYSTEM_ERROR("cannot make processx socketpair");
+    }
   }
 
   processx__cloexec_fcntl(pipe[0], 1);
@@ -264,10 +414,10 @@ skip:
 }
 
 SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
-		   SEXP std_err, SEXP connections, SEXP env,
-		   SEXP windows_verbatim_args, SEXP windows_hide_window,
-		   SEXP private, SEXP cleanup,  SEXP wd, SEXP encoding,
-		   SEXP tree_id) {
+		   SEXP std_err, SEXP pty, SEXP pty_options,
+                   SEXP connections, SEXP env, SEXP windows_verbatim_args,
+                   SEXP windows_hide_window, SEXP private, SEXP cleanup,
+                   SEXP wd, SEXP encoding, SEXP tree_id) {
 
   char *ccommand = processx__tmp_string(command, 0);
   char **cargs = processx__tmp_character(args);
@@ -276,6 +426,7 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
   const char *cstdin = isNull(std_in) ? 0 : CHAR(STRING_ELT(std_in, 0));
   const char *cstdout = isNull(std_out) ? 0 : CHAR(STRING_ELT(std_out, 0));
   const char *cstderr = isNull(std_err) ? 0 : CHAR(STRING_ELT(std_err, 0));
+  const int cpty = LOGICAL(pty)[0];
   const char *cencoding = CHAR(STRING_ELT(encoding, 0));
   const char *ctree_id = CHAR(STRING_ELT(tree_id, 0));
   processx_options_t options = { 0 };
@@ -287,6 +438,10 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
   int signal_pipe[2] = { -1, -1 };
   int (*pipes)[2];
   int i;
+  int pty_master_fd;
+#define R_PROCESSX_PTY_NAME_LEN 2014
+  char pty_namex[R_PROCESSX_PTY_NAME_LEN];
+  char *pty_name = cpty ? pty_namex : 0;
 
   processx_handle_t *handle = NULL;
   SEXP result;
@@ -297,7 +452,7 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
   options.wd = isNull(wd) ? 0 : CHAR(STRING_ELT(wd, 0));
 
   if (pipe(signal_pipe)) {
-    PROCESSX__ERROR("Cannot create pipe", strerror(errno));
+    R_THROW_SYSTEM_ERROR("Cannot create pipe when running '%s'", ccommand);
   }
   processx__cloexec_fcntl(signal_pipe[0], 1);
   processx__cloexec_fcntl(signal_pipe[1], 1);
@@ -307,16 +462,25 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
   result = PROTECT(processx__make_handle(private, ccleanup));
   handle = R_ExternalPtrAddr(result);
 
-  /* Create pipes, if requested. */
-  if (cstdin && !strcmp(cstdin, "|")) processx__make_socketpair(pipes[0]);
-  if (cstdout && !strcmp(cstdout, "|")) processx__make_socketpair(pipes[1]);
-  if (cstderr && !strcmp(cstderr, "|")) processx__make_socketpair(pipes[2]);
+  if (cpty) {
+    pty_master_fd =
+      processx__pty_master_open(pty_name, R_PROCESSX_PTY_NAME_LEN);
+    if (pty_master_fd == -1) {
+      R_THROW_SYSTEM_ERROR("Cannot open pty when running '%s'", ccommand);
+    }
+    options.pty_echo = LOGICAL(VECTOR_ELT(pty_options, 0))[0];
+
+  } else {
+    /* Create pipes, if requested. */
+    if (cstdin && !strcmp(cstdin, "|")) processx__make_socketpair(pipes[0], ccommand);
+    if (cstdout && !strcmp(cstdout, "|")) processx__make_socketpair(pipes[1], ccommand);
+    if (cstderr && !strcmp(cstderr, "|")) processx__make_socketpair(pipes[2], ccommand);
+  }
 
   for (i = 0; i < num_connections - 3; i++) {
     processx_connection_t *ccon =
       R_ExternalPtrAddr(VECTOR_ELT(connections, i));
     int fd = processx_c_connection_fileno(ccon);
-    processx__nonblock_fcntl(fd, 0);
     pipes[i + 3][1] = fd;
   }
 
@@ -329,17 +493,22 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
     err = -errno;
     if (signal_pipe[0] >= 0) close(signal_pipe[0]);
     if (signal_pipe[1] >= 0) close(signal_pipe[1]);
+    if (cpty) close(pty_master_fd);
     processx__unblock_sigchld();
-    PROCESSX__ERROR("Cannot fork", strerror(err));
+    R_THROW_SYSTEM_ERROR_CODE(err, "Cannot fork when running '%s'",
+                              ccommand);
   }
 
   /* CHILD */
   if (pid == 0) {
     /* LCOV_EXCL_START */
+    if (cpty) close(pty_master_fd);
+    processx__unblock_sigchld();
     processx__child_init(handle, pipes, num_connections, ccommand, cargs,
-			 signal_pipe[1], cstdin, cstdout, cstderr, cenv,
-			 &options, ctree_id);
-    PROCESSX__ERROR("Cannot start child process", "");
+			 signal_pipe[1], cstdin, cstdout, cstderr,
+                         pty_name, cenv, &options, ctree_id);
+    R_THROW_SYSTEM_ERROR("Cannot start child process when running '%s'",
+                         ccommand);
     /* LCOV_EXCL_STOP */
   }
 
@@ -347,13 +516,17 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
      to avoid race conditions when sending signals */
   handle->create_time = processx__create_time(pid);
 
+  handle->ptyfd = -1;
+  if (cpty) handle->ptyfd = pty_master_fd;
+
   /* We need to know the processx children */
   if (processx__child_add(pid, result)) {
     err = -errno;
     if (signal_pipe[0] >= 0) close(signal_pipe[0]);
     if (signal_pipe[1] >= 0) close(signal_pipe[1]);
     processx__unblock_sigchld();
-    PROCESSX__ERROR("Cannot create child process", "out of memory");
+    R_THROW_ERROR("Cannot create child process '%s', out of memory",
+                  ccommand);
   }
 
   /* SIGCHLD can arrive now */
@@ -378,7 +551,9 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
     } while (err == -1 && errno == EINTR);
 
   } else {
-    PROCESSX__ERROR("Child process failed to start", strerror(exec_errorno));
+    R_THROW_SYSTEM_ERROR_CODE(-exec_errorno,
+                              "Child process '%s' failed to start",
+                              ccommand);
   }
 
   if (signal_pipe[0] >= 0) close(signal_pipe[0]);
@@ -403,13 +578,6 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
     if (pipes[i][1] >= 0) close(pipes[i][1]);
   }
 
-  /* Close connections passed to the child process */
-  for (i = 3; i < num_connections; i++) {
-    processx_connection_t *ccon =
-      R_ExternalPtrAddr(VECTOR_ELT(connections, i - 3));
-    processx_c_connection_close(ccon);
-  }
-
   /* Create proper connections */
   processx__create_connections(handle, private, cencoding);
 
@@ -419,8 +587,10 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
     return result;
   }
 
-  error("processx error: '%s' at %s:%d", strerror(- exec_errorno),
-	__FILE__, __LINE__);
+  R_THROW_SYSTEM_ERROR_CODE(-exec_errorno,
+                            "cannot start processx process '%s'",
+                            ccommand);
+  return R_NilValue;
 }
 
 void processx__collect_exit_status(SEXP status, int retval, int wstat) {
@@ -430,7 +600,7 @@ void processx__collect_exit_status(SEXP status, int retval, int wstat) {
      So we are not blocking it here. */
 
   if (!handle) {
-    error("Invalid handle, already finalized");
+    R_THROW_ERROR("Invalid handle, already finalized");
   }
 
   if (handle->collected) { return; }
@@ -448,6 +618,14 @@ void processx__collect_exit_status(SEXP status, int retval, int wstat) {
   handle->collected = 1;
 }
 
+static void processx__wait_cleanup(void *ptr) {
+  int *fds = (int*) ptr;
+  if (!fds) return;
+  if (fds[0] >= 0) close(fds[0]);
+  if (fds[1] >= 0) close(fds[1]);
+  free(fds);
+}
+
 /* In general we need to worry about three asynchronous processes here:
  * 1. The main code, i.e. the code in this function.
  * 2. The finalizer, that can be triggered by any R function.
@@ -473,18 +651,24 @@ void processx__collect_exit_status(SEXP status, int retval, int wstat) {
  * 7. We keep polling until the timeout expires or the process finishes.
  */
 
-SEXP processx_wait(SEXP status, SEXP timeout) {
+SEXP processx_wait(SEXP status, SEXP timeout, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   int ctimeout = INTEGER(timeout)[0], timeleft = ctimeout;
   struct pollfd fd;
   int ret = 0;
   pid_t pid;
 
+  int *fds = malloc(sizeof(int) * 2);
+  if (!fds) R_THROW_SYSTEM_ERROR("Allocating memory when waiting");
+  fds[0] = fds[1] = -1;
+  r_call_on_exit(processx__wait_cleanup, fds);
+
   processx__block_sigchld();
 
   if (!handle) {
     processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
+    return ScalarLogical(1);
   }
 
   pid = handle->pid;
@@ -502,8 +686,10 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
   /* Setup the self-pipe that we can poll */
   if (pipe(handle->waitpipe)) {
     processx__unblock_sigchld();
-    error("processx error: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("processx error when waiting for '%s'", cname);
   }
+  fds[0] = handle->waitpipe[0];
+  fds[1] = handle->waitpipe[1];
   processx__nonblock_fcntl(handle->waitpipe[0], 1);
   processx__nonblock_fcntl(handle->waitpipe[1], 1);
 
@@ -514,6 +700,8 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
 
   processx__unblock_sigchld();
 
+
+
   while (ctimeout < 0 || timeleft > PROCESSX_INTERRUPT_INTERVAL) {
     do {
       ret = poll(&fd, 1, PROCESSX_INTERRUPT_INTERVAL);
@@ -545,12 +733,12 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
   }
 
   if (ret == -1) {
-    error("processx wait with timeout error: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("processx wait with timeout error while "
+                         "waiting for '%s'", cname);
   }
 
  cleanup:
-  if (handle->waitpipe[0] >= 0) close(handle->waitpipe[0]);
-  if (handle->waitpipe[1] >= 0) close(handle->waitpipe[1]);
+  /* pipe is closed in the on_exit handler */
   handle->waitpipe[0] = -1;
   handle->waitpipe[1] = -1;
 
@@ -572,19 +760,16 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
  * 6. Otherwise we collect the exit status, and return FALSE.
  */
 
-SEXP processx_is_alive(SEXP status) {
+SEXP processx_is_alive(SEXP status, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   pid_t pid;
   int wstat, wp;
   int ret = 0;
 
   processx__block_sigchld();
 
-  if (!handle) {
-    processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
-  }
-
+  if (!handle) goto cleanup;
   if (handle->collected) goto cleanup;
 
   /* Otherwise a non-blocking waitpid to collect zombies */
@@ -603,7 +788,7 @@ SEXP processx_is_alive(SEXP status) {
   /* Some other error? */
   if (wp == -1) {
     processx__unblock_sigchld();
-    error("processx_is_alive: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("processx_is_alive, process '%s'", cname);
   }
 
   /* If running, return TRUE, otherwise collect exit status, return FALSE */
@@ -622,8 +807,9 @@ SEXP processx_is_alive(SEXP status) {
  * exit status if the process has already finished. See above.
  */
 
-SEXP processx_get_exit_status(SEXP status) {
+SEXP processx_get_exit_status(SEXP status, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   pid_t pid;
   int wstat, wp;
   SEXP result;
@@ -631,8 +817,8 @@ SEXP processx_get_exit_status(SEXP status) {
   processx__block_sigchld();
 
   if (!handle) {
-    processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
+    result = PROTECT(ScalarInteger(NA_INTEGER));
+    goto cleanup;
   }
 
   /* If we already have the status, then just return */
@@ -658,7 +844,7 @@ SEXP processx_get_exit_status(SEXP status) {
   /* Some other error? */
   if (wp == -1) {
     processx__unblock_sigchld();
-    error("processx_get_exit_status: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("processx_get_exit_status error for '%s'", cname);
   }
 
   /* If running, do nothing otherwise collect */
@@ -686,16 +872,17 @@ SEXP processx_get_exit_status(SEXP status) {
  * the SIGCHLD handler.
  */
 
-SEXP processx_signal(SEXP status, SEXP signal) {
+SEXP processx_signal(SEXP status, SEXP signal, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   pid_t pid;
   int wstat, wp, ret, result;
 
   processx__block_sigchld();
 
   if (!handle) {
-    processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
+    result = 0;
+    goto cleanup;
   }
 
   /* If we already have the status, then return `FALSE` */
@@ -714,7 +901,7 @@ SEXP processx_signal(SEXP status, SEXP signal) {
     result = 0;
   } else {
     processx__unblock_sigchld();
-    error("processx_signal: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("processx_signal for '%s'", cname);
     return R_NilValue;
   }
 
@@ -731,7 +918,7 @@ SEXP processx_signal(SEXP status, SEXP signal) {
 
   if (wp == -1) {
     processx__unblock_sigchld();
-    error("processx_signal: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("processx_signal for '%s'", cname);
   }
 
  cleanup:
@@ -739,8 +926,8 @@ SEXP processx_signal(SEXP status, SEXP signal) {
   return ScalarLogical(result);
 }
 
-SEXP processx_interrupt(SEXP status) {
-  return processx_signal(status, ScalarInteger(2));
+SEXP processx_interrupt(SEXP status, SEXP name) {
+  return processx_signal(status, ScalarInteger(2), name);
 }
 
 /* This is a special case of `processx_signal`, and we implement it almost
@@ -754,17 +941,15 @@ SEXP processx_interrupt(SEXP status) {
  * still alive or not.
  */
 
-SEXP processx_kill(SEXP status, SEXP grace) {
+SEXP processx_kill(SEXP status, SEXP grace, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   pid_t pid;
   int wstat, wp, result = 0;
 
   processx__block_sigchld();
 
-  if (!handle) {
-    processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
-  }
+  if (!handle) { goto cleanup; }
 
   /* Check if we have an exit status, it yes, just return (FALSE) */
   if (handle->collected) { goto cleanup; }
@@ -785,7 +970,7 @@ SEXP processx_kill(SEXP status, SEXP grace) {
   /* Some other error? */
   if (wp == -1) {
     processx__unblock_sigchld();
-    error("processx_kill: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("processx_kill for '%s'", cname);
   }
 
   /* If the process is not running, return (FALSE) */
@@ -796,7 +981,7 @@ SEXP processx_kill(SEXP status, SEXP grace) {
   if (ret == -1 && (errno == ESRCH || errno == EPERM)) { goto cleanup; }
   if (ret == -1) {
     processx__unblock_sigchld();
-    error("process_kill: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("process_kill for '%s'", cname);
   }
 
   /* Do a waitpid to collect the status and reap the zombie */
@@ -820,7 +1005,9 @@ SEXP processx_kill(SEXP status, SEXP grace) {
 SEXP processx_get_pid(SEXP status) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
 
-  if (!handle) { error("Internal processx error, handle already removed"); }
+  /* This might happen if it was finalized at the end of the session,
+     even though there are some references to the R object. */
+  if (!handle) return ScalarInteger(NA_INTEGER);
 
   return ScalarInteger(handle->pid);
 }
@@ -837,7 +1024,7 @@ SEXP processx__process_exists(SEXP pid) {
   } else if (errno == ESRCH) {
     return ScalarLogical(0);
   } else {
-    error("kill syscall error: %s", strerror(errno));
+    R_THROW_SYSTEM_ERROR("kill syscall error for pid '%d'", cpid);
     return R_NilValue;
   }
 }
